/*
 * Copyright 2008-2009 Michael Lotz <mmlr@mlotz.ch>
 * Distributed under the terms of the MIT license.
 */


//!	Driver for USB Human Interface Devices.


#include "Driver.h"
#include "GamepadDevice.h"

#include "HIDDevice.h"
#include "HIDReport.h"
#include "HIDReportItem.h"
#include "HIDCollection.h"

#include <new>
#include <string.h>
#include <usb/USB_hid.h>

#include <keyboard_mouse_driver.h>

#define JOYSTICK_READ 'jrea'

GamepadDevice::GamepadDevice(HIDReport *report, HIDReportItem *xAxis,
	HIDReportItem *yAxis)
	:
	ProtocolHandler(report->Device(), "joystick/usb/", 512),
	fReport(report),
	fXAxis(xAxis),
	fYAxis(yAxis),
	fNumButtons(0),
	fLastClickTime(0),
	fClickSpeed(250000)
{
	fNumAxes = 2;
	
	for (uint32 i = 0; i < report->CountItems(); i++) {
		HIDReportItem *item = report->ItemAt(i);
		
		TRACE_ALWAYS("GAMEPAD USAGES: UsagePage %d UsageId %d\n", item->UsagePage(), item->UsageID());
		if (!item->HasData())
			continue;

		if (item->UsagePage() == B_HID_USAGE_PAGE_BUTTON
			&& item->UsageID() - 1 < B_MAX_JOYSTICK_BUTTONS)
			fButtons[fNumButtons++] = item;
	}

	fButtons[fNumButtons] = NULL;
	memset(&fJoyModInfo, 0, sizeof(joystick_module_info));
	
	TRACE_ALWAYS("gamepad device with %lu buttons\n", fNumButtons);
	TRACE_ALWAYS("report id: %u\n", report->ID());
}


ProtocolHandler *
GamepadDevice::AddHandler(HIDDevice *device, HIDReport *report)
{
	TRACE_ALWAYS("GAMEPAD latest\n");
	
	HIDCollection *collection = device->Parser()->RootCollection();
	collection->PrintToStream();
	
	 //check to see if we have a GamePad Device
	if(collection->UsagePage() != B_HID_USAGE_PAGE_GENERIC_DESKTOP
		|| collection->UsageID() != B_HID_UID_GD_GAMEPAD) return NULL;

#if 0
	HIDReportItem *game = report->FindItem(B_HID_USAGE_PAGE_GENERIC_DESKTOP,
		B_HID_UID_GD_GAMEPAD);
		
	
	TRACE_ALWAYS("gamepad device is null %s\n",
		game == NULL ? "yes " : "no ");
#endif

	//if(game == NULL)
	//	return NULL;
	//TRACE("GAMEPAD USAGES: UsagePage %d UsageId %d\n", collection->UsagePage(), collection->UsageID());

	// try to find at least an x and y axis
	HIDReportItem *xAxis = report->FindItem(B_HID_USAGE_PAGE_GENERIC_DESKTOP,
		B_HID_UID_GD_X);
	if (xAxis == NULL)
		return NULL;

	HIDReportItem *yAxis = report->FindItem(B_HID_USAGE_PAGE_GENERIC_DESKTOP,
		B_HID_UID_GD_Y);
	if (yAxis == NULL)
		return NULL;

	return new(std::nothrow) GamepadDevice(report, xAxis, yAxis);
}


status_t
GamepadDevice::Control(uint32 *cookie, uint32 op, void *buffer, size_t length)
{
	TRACE("IN GAMEPADDEVICE_CONTROL\n");
	switch (op) {
		case B_JOYSTICK_SET_DEVICE_MODULE:
		{
			fJoyModInfo = *(joystick_module_info *)buffer;
			fJoyModInfo.num_axes = NumAxes();
			fJoyModInfo.num_buttons = NumButtons();;
			TRACE("B_JOYSTICK_SET_DEVICE_MODULE\n");
		} return B_OK;
		case B_JOYSTICK_GET_DEVICE_MODULE:
			*(joystick_module_info *)buffer = fJoyModInfo;
			TRACE("B_JOYSTICK_GET_DEVICE_MODULE\n");
			return B_OK;
		
		case JOYSTICK_READ:
			while (RingBufferReadable() == 0) {
				status_t result = _ReadReport();
				if (result != B_OK)
					return result;
			}

			return RingBufferRead(buffer, sizeof(extended_joystick));

		case MS_NUM_EVENTS:
		{
			int32 count = RingBufferReadable() / sizeof(mouse_movement);
			if (count == 0 && fReport->Device()->IsRemoved())
				return B_DEV_NOT_READY;
			return count;
		}

		case MS_SET_CLICKSPEED:
#ifdef __HAIKU__
				return user_memcpy(&fClickSpeed, buffer, sizeof(bigtime_t));
#else
				fClickSpeed = *(bigtime_t *)buffer;
				return B_OK;
#endif
	}

	return B_ERROR;
}


uint32
GamepadDevice::NumButtons()
{
	return fNumButtons;
}


uint32
GamepadDevice::NumAxes()
{
	return fNumAxes;
}

	
status_t
GamepadDevice::_ReadReport()
{
	status_t result = fReport->WaitForReport(B_INFINITE_TIMEOUT);
	if (result != B_OK) {
		if (fReport->Device()->IsRemoved()) {
			TRACE("device has been removed\n");
			return B_DEV_NOT_READY;
		}

		if (result != B_INTERRUPTED) {
			// interrupts happen when other reports come in on the same
			// input as ours
			TRACE_ALWAYS("error waiting for report: %s\n", strerror(result));
		}

		// signal that we simply want to try again
		return B_OK;
	}

	extended_joystick info;
	memset(&info, 0, sizeof(info));

	if (fXAxis->Extract() == B_OK && fXAxis->Valid())
		info.axes[0] = fXAxis->Data();
	if (fYAxis->Extract() == B_OK && fYAxis->Valid())
		info.axes[1] = fYAxis->Data();

	for (uint32 i = 0; i < B_MAX_JOYSTICK_BUTTONS; i++) {
		HIDReportItem *button = fButtons[i];
		if (button == NULL)
			break;

		if (button->Extract() == B_OK && button->Valid())
			info.buttons |= (button->Data() & 1) << (button->UsageID() - 1);
	}

	fReport->DoneProcessing();
	TRACE("got joystick report\n");

	bigtime_t timestamp = system_time();
	
	#if 0
	if (info.buttons != 0) {
		if (fLastButtons == 0) {
			if (fLastClickTime + fClickSpeed > timestamp)
				fClickCount++;
			else
				fClickCount = 1;
		}

		fLastClickTime = timestamp;
		info.clicks = fClickCount;
	}
	#endif
	
	//fLastButtons = info.buttons;
	info.timestamp = timestamp;
	return RingBufferWrite(&info, sizeof(extended_joystick));
}
